@@ -1,17 +1,15 @@
 # -*- coding: utf-8 -*-
-import socket
 from functools import partial
 from itertools import izip
 import logging
 import weakref
-from collections import deque
 
 from tornado.ioloop import IOLoop
-from tornado.iostream import IOStream
 from tornado import gen
 
 from datetime import datetime
-from exceptions import RequestError, ConnectionError, ResponseError
+from .exceptions import RequestError, ConnectionError, ResponseError
+from .connection import Connection
 
 
 log = logging.getLogger('tornadoredis.client')
@@ -50,165 +48,11 @@ def string_keys_to_dict(key_string, callback):
 
 def dict_merge(*dicts):
     merged = {}
-    [merged.update(d) for d in dicts]
+    for d in dicts:
+        merged.update(d)
     return merged
 
 
-def encode(value):
-    if isinstance(value, str):
-        return value
-    elif isinstance(value, unicode):
-        return value.encode('utf-8')
-    # pray and hope
-    return str(value)
-
-
-def format_command(*tokens, **kwargs):
-    cmds = []
-    for t in tokens:
-        e_t = encode(t)
-        cmds.append('$%s\r\n%s\r\n' % (len(e_t), e_t))
-    return '*%s\r\n%s' % (len(tokens), ''.join(cmds))
-
-
-def format_pipeline_request(command_stack):
-    return ''.join(format_command(c.cmd, *c.args, **c.kwargs) \
-                   for c in command_stack)
-
-
-class Connection(object):
-    def __init__(self, host, port, event_handler,
-                 stop_after=None, io_loop=None):
-        self.host = host
-        self.port = port
-        self._event_handler = weakref.proxy(event_handler)
-        self.timeout = stop_after
-        self._stream = None
-        self._io_loop = io_loop
-        self.try_left = 2
-
-        self.in_progress = False
-        self.read_callbacks = []
-        self.ready_callbacks = deque()
-
-    def __del__(self):
-        self.disconnect()
-
-    def __enter__(self):
-        return self
-
-    def __exit__(self, *args, **kwargs):
-        if self.ready_callbacks:
-            # Pop a SINGLE callback from the queue and execute it.
-            # The next one will be executed from the code
-            # invoked by the callback
-            callback = self.ready_callbacks.popleft()
-            callback()
-
-    def ready(self):
-        return not self.read_callbacks and not self.ready_callbacks
-
-    def wait_until_ready(self, callback=None):
-        if callback:
-            if not self.ready():
-                self.ready_callbacks.append(callback)
-            else:
-                callback()
-        return self
-
-    def connect(self):
-        if not self._stream:
-            try:
-                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
-                sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
-                sock.settimeout(self.timeout)
-                sock.connect((self.host, self.port))
-                self._stream = IOStream(sock, io_loop=self._io_loop)
-                self._stream.set_close_callback(self.on_stream_close)
-                self.connected()
-            except socket.error, e:
-                raise ConnectionError(str(e))
-            self.fire_event('on_connect')
-
-    def on_stream_close(self):
-        if self._stream:
-            self._stream = None
-            callbacks = self.read_callbacks
-            self.read_callbacks = []
-            for callback in callbacks:
-                callback(None)
-
-    def disconnect(self):
-        if self._stream:
-            s = self._stream
-            self._stream = None
-            try:
-                s.socket.shutdown(socket.SHUT_RDWR)
-                s.close()
-            except socket.error:
-                pass
-
-    def fire_event(self, event):
-        if self._event_handler:
-            try:
-                getattr(self._event_handler, event)()
-            except AttributeError:
-                pass
-
-    def write(self, data, try_left=None):
-        if try_left is None:
-            try_left = self.try_left
-        if not self._stream:
-            self.connect()
-            if not self._stream:
-                raise ConnectionError('Tried to write to '
-                                      'non-existent connection')
-
-        if try_left > 0:
-            try:
-                self._stream.write(data)
-            except IOError:
-                self.disconnect()
-                self.write(data, try_left - 1)
-        else:
-            raise ConnectionError('Tried to write to non-existent connection')
-
-    def read(self, length, callback=None):
-        try:
-            if not self._stream:
-                self.disconnect()
-                raise ConnectionError('Tried to read from '
-                                      'non-existent connection')
-            self.read_callbacks.append(callback)
-            self._stream.read_bytes(length,
-                                    callback=partial(self.read_callback,
-                                                     callback))
-        except IOError:
-            self.fire_event('on_disconnect')
-
-    def read_callback(self, callback, *args, **kwargs):
-        self.read_callbacks.remove(callback)
-        callback(*args, **kwargs)
-
-    def readline(self, callback=None):
-        try:
-            if not self._stream:
-                self.disconnect()
-                raise ConnectionError('Tried to read from '
-                                      'non-existent connection')
-            self.read_callbacks.append(callback)
-            self._stream.read_until('\r\n',
-                                    callback=partial(self.read_callback,
-                                                     callback))
-        except IOError:
-            self.fire_event('on_disconnect')
-
-    def connected(self):
-        if self._stream:
-            return True
-        return False
-
-
 def reply_to_bool(r, *args, **kwargs):
     return bool(r)
 
@@ -261,7 +105,8 @@ def reply_info(response):
     info = {}
 
     def get_value(value):
-        if ',' not in value:
+        # Does this string contains subvalues?
+        if (',' not in value) or ('=' not in value):
             return value
         sub_dict = {}
         for item in value.split(','):
@@ -272,11 +117,13 @@ def reply_info(response):
                 sub_dict[k] = v
         return sub_dict
     for line in response.splitlines():
-        key, value = line.split(':')
-        try:
-            info[key] = int(value)
-        except ValueError:
-            info[key] = get_value(value)
+        line = line.strip()
+        if line and not line.startswith('#'):
+            key, value = line.split(':')
+            try:
+                info[key] = int(value)
+            except ValueError:
+                info[key] = get_value(value)
     return info
 
 
@@ -284,6 +131,13 @@ def reply_ttl(r, *args, **kwargs):
     return r != -1 and r or None
 
 
+def to_list(source):
+    if isinstance(source, basestring):
+        return [source]
+    else:
+        return list(source)
+
+
 PUB_SUB_COMMANDS = set([
     'SUBSCRIBE',
     'PSUBSCRIBE',
@@ -294,52 +148,77 @@ PUB_SUB_COMMANDS = set([
 ])
 
 
+REPLY_MAP = dict_merge(
+    string_keys_to_dict('AUTH BGREWRITEAOF BGSAVE DEL EXISTS '
+                        'EXPIRE HDEL HEXISTS '
+                        'HMSET MOVE MSET MSETNX SAVE SETNX',
+                        reply_to_bool),
+    string_keys_to_dict('FLUSHALL FLUSHDB SELECT SET SETEX '
+                        'SHUTDOWN RENAME RENAMENX WATCH UNWATCH',
+                        make_reply_assert_msg('OK')),
+    string_keys_to_dict('SMEMBERS SINTER SUNION SDIFF',
+                        reply_set),
+    string_keys_to_dict('HGETALL BRPOP BLPOP',
+                        reply_dict_from_pairs),
+    string_keys_to_dict('HGET',
+                        reply_str),
+    string_keys_to_dict('SUBSCRIBE UNSUBSCRIBE LISTEN '
+                        'PSUBSCRIBE UNSUBSCRIBE',
+                        reply_pubsub_message),
+    string_keys_to_dict('ZRANK ZREVRANK',
+                        reply_int),
+    string_keys_to_dict('ZSCORE ZINCRBY ZCOUNT ZCARD',
+                        reply_int),
+    string_keys_to_dict('ZRANGE ZRANGEBYSCORE ZREVRANGE '
+                        'ZREVRANGEBYSCORE',
+                        reply_zset),
+    {'HMGET': reply_hmget},
+    {'PING': make_reply_assert_msg('PONG')},
+    {'LASTSAVE': reply_datetime},
+    {'TTL': reply_ttl},
+    {'INFO': reply_info},
+    {'MULTI_PART': make_reply_assert_msg('QUEUED')},
+)
+
+
 class Client(object):
+#    __slots__ = ('_io_loop', '_connection_pool', 'connection', 'subscribed',
+#                 'password', 'selected_db', '_pipeline', '_weak')
+
     def __init__(self, host='localhost', port=6379, password=None,
-            selected_db=None, io_loop=None):
+                 selected_db=None, io_loop=None, connection_pool=None):
         self._io_loop = io_loop or IOLoop.instance()
-        self.connection = Connection(host, port, self, io_loop=self._io_loop)
-        self.current_cmd_line = None
+        self._connection_pool = connection_pool
+        self._weak = weakref.proxy(self)
+        if connection_pool:
+            connection = (connection_pool
+                          .get_connection(event_handler_proxy=self._weak))
+        else:
+            connection = Connection(host=host, port=port,
+                                    weak_event_handler=self._weak,
+                                    io_loop=self._io_loop)
+        self.connection = connection
         self.subscribed = False
         self.password = password
         self.selected_db = selected_db
-        self.REPLY_MAP = dict_merge(
-                string_keys_to_dict('AUTH BGREWRITEAOF BGSAVE DEL EXISTS '
-                                    'EXPIRE HDEL HEXISTS '
-                                    'HMSET MOVE MSET MSETNX SAVE SETNX',
-                                    reply_to_bool),
-                string_keys_to_dict('FLUSHALL FLUSHDB SELECT SET SETEX '
-                                    'SHUTDOWN RENAME RENAMENX WATCH UNWATCH',
-                                    make_reply_assert_msg('OK')),
-                string_keys_to_dict('SMEMBERS SINTER SUNION SDIFF',
-                                    reply_set),
-                string_keys_to_dict('HGETALL BRPOP BLPOP',
-                                    reply_dict_from_pairs),
-                string_keys_to_dict('HGET',
-                                    reply_str),
-                string_keys_to_dict('SUBSCRIBE UNSUBSCRIBE LISTEN '
-                                    'PSUBSCRIBE UNSUBSCRIBE',
-                                    reply_pubsub_message),
-                string_keys_to_dict('ZRANK ZREVRANK',
-                                    reply_int),
-                string_keys_to_dict('ZSCORE ZINCRBY ZCOUNT ZCARD',
-                                    reply_int),
-                string_keys_to_dict('ZRANGE ZRANGEBYSCORE ZREVRANGE '
-                                    'ZREVRANGEBYSCORE',
-                                    reply_zset),
-                {'HMGET': reply_hmget},
-                {'PING': make_reply_assert_msg('PONG')},
-                {'LASTSAVE': reply_datetime},
-                {'TTL': reply_ttl},
-                {'INFO': reply_info},
-                {'MULTI_PART': make_reply_assert_msg('QUEUED')},
-            )
-
         self._pipeline = None
 
+    def __del__(self):
+        try:
+            connection = self.connection
+            pool = self._connection_pool
+        except AttributeError:
+            connection = None
+            pool = None
+        if connection:
+            if pool:
+                pool.release(connection)
+                connection.wait_until_ready()
+            else:
+                connection.disconnect()
+
     def __repr__(self):
-        return 'Tornadoredis client (host=%s, port=%s)' \
-                % (self.connection.host, self.connection.port)
+        return 'tornadoredis.Client (db=%s)' % (self.selected_db)
 
     def __enter__(self):
         return self
@@ -347,7 +226,44 @@ class Client(object):
     def __exit__(self, *args, **kwargs):
         pass
 
+    def __getattribute__(self, item):
+        '''
+        Bind methods to the weak proxy to avoid memory leaks
+        when bound method is passed as argument to the gen.Task
+        constructor.
+        '''
+        a = super(Client, self).__getattribute__(item)
+        try:
+            if callable(a) and a.__self__:
+                try:
+                    a = self.__class__.__dict__[item]
+                except KeyError:
+                    a = Client.__dict__[item]
+                a = partial(a, self._weak)
+        except AttributeError:
+            pass
+        return a
+
     def pipeline(self, transactional=False):
+        '''
+        Create the 'Pipeline' to pack multiple redis commands
+        in a single request.
+
+        Usage:
+            pipe = self.client.pipeline()
+            pipe.hset('foo', 'bar', 1)
+            pipe.expire('foo', 60)
+
+            yield gen.Task(pipe.execute)
+
+        or:
+
+            with self.client.pipeline() as pipe:
+                pipe.hset('foo', 'bar', 1)
+                pipe.expire('foo', 60)
+
+                yield gen.Task(pipe.execute)
+        '''
         if not self._pipeline:
             self._pipeline = Pipeline(
                 selected_db=self.selected_db,
@@ -357,25 +273,53 @@ class Client(object):
             self._pipeline.connection = self.connection
         return self._pipeline
 
-    #### connection
-
-    def connect(self):
-        self.connection.connect()
-
-    def disconnect(self):
-        self.connection.disconnect()
-
+    #### Connection event handlers
     def on_connect(self):
         if self.password:
-            self.auth(self.password)
+            password = self.connection.info.get('password', None)
+            if password != self.password:
+                self.auth(self.password)
+                self.connection.info['password'] = self.password
         if self.selected_db:
-            self.select(self.selected_db)
+            db = self.connection.info.get('db', None)
+            if db != self.selected_db:
+                self.select(self.selected_db)
+                self.connection.info['db'] = self.selected_db
 
     def on_disconnect(self):
         if self.subscribed:
             self.subscribed = False
         raise ConnectionError("Socket closed on remote end")
-    ####
+
+    #### connection
+    def connect(self):
+        if not self.connection.connected():
+            pool = self._connection_pool
+            if pool:
+                old_conn = self.connection
+                self.connection = pool.get_connection(event_handler=self)
+                self.connection.ready_callbacks = old_conn.ready_callbacks
+            else:
+                self.connection.connect()
+
+    @gen.engine
+    def disconnect(self, callback=None):
+        '''
+        Disconnect from the Redis server.
+        '''
+        connection = self.connection
+        if connection:
+            pool = self._connection_pool
+            if pool:
+                pool.release(connection)
+                yield gen.Task(connection.wait_until_ready)
+                proxy = pool.make_proxy(client_proxy=self._weak,
+                                        connected=False)
+                self.connection = proxy
+            else:
+                self.connection.disconnect()
+        if callback:
+            callback(False)
 
     #### formatting
     def encode(self, value):
@@ -383,10 +327,9 @@ class Client(object):
             return value
         elif isinstance(value, unicode):
             return value.encode('utf-8')
-        # pray and hope
         return str(value)
 
-    def format_command(self, *tokens):
+    def format_command(self, *tokens, **kwargs):
         cmds = []
         for t in tokens:
             e_t = self.encode(t)
@@ -394,12 +337,12 @@ class Client(object):
         return '*%s\r\n%s' % (len(tokens), ''.join(cmds))
 
     def format_reply(self, cmd_line, data):
-        if cmd_line.cmd not in self.REPLY_MAP:
+        if cmd_line.cmd not in REPLY_MAP:
             return data
         try:
-            res = self.REPLY_MAP[cmd_line.cmd](data,
-                                               *cmd_line.args,
-                                               **cmd_line.kwargs)
+            res = REPLY_MAP[cmd_line.cmd](data,
+                                          *cmd_line.args,
+                                          **cmd_line.kwargs)
         except Exception, e:
             raise ResponseError(
                 'failed to format reply to %s, raw data: %s; err message: %s'
@@ -408,11 +351,6 @@ class Client(object):
         return res
     ####
 
-    #### new AsIO
-    def call_callback(self, callback, *args, **kwargs):
-        for cb in callback:
-            cb(*args, **kwargs)
-
     @gen.engine
     def execute_command(self, cmd, *args, **kwargs):
         callback = kwargs.get('callback', None)
@@ -424,29 +362,43 @@ class Client(object):
                 cmd_line))
             return
 
-        if not self.connection.connected():
-            self.connection.connect()
+        n_tries = 2
+        while n_tries > 0:
+            n_tries -= 1
+            if not self.connection.connected():
+                self.connection.connect()
 
-        if not self.connection.ready() and not self.subscribed:
-            yield gen.Task(self.connection.wait_until_ready)
+            if not self.connection.ready() and not self.subscribed:
+                yield gen.Task(self.connection.wait_until_ready)
 
-        try:
-            self.connection.write(self.format_command(cmd, *args, **kwargs))
-        except Exception, e:
-            self.connection.disconnect()
-            raise e
-
-        if (cmd in PUB_SUB_COMMANDS) or (self.subscribed and cmd == 'PUBLISH'):
-            result = True
-        else:
             with self.connection as c:
-                result = None
-                data = yield gen.Task(c.readline)
-                if not data:
-                    raise ConnectionError('no data received')
+                command = self.format_command(cmd, *args, **kwargs)
+                try:
+                    yield gen.Task(self.connection.write, command)
+                except Exception, e:
+                    self.connection.disconnect()
+                    if not n_tries:
+                        raise e
+                    else:
+                        continue
+
+                if ((cmd in PUB_SUB_COMMANDS) or
+                    (self.subscribed and cmd == 'PUBLISH')):
+                    result = True
+                    break
                 else:
-                    resp = yield gen.Task(self.process_data, data, cmd_line)
-                    result = self.format_reply(cmd_line, resp)
+                    result = None
+                    data = yield gen.Task(c.readline)
+                    if not data:
+                        if not n_tries:
+                            raise ConnectionError('no data received')
+                    else:
+                        resp = yield gen.Task(self.process_data,
+                                              data,
+                                              cmd_line)
+                        result = self.format_reply(cmd_line, resp)
+                        break
+
         if callback:
             callback(result)
 
@@ -459,8 +411,6 @@ class Client(object):
         elif data == '*0' or data == '*-1':
             response = []
         else:
-            if len(data) == 0:
-                raise IOError('Disconnected')
             head, tail = data[0], data[1:]
 
             if head == '*':
@@ -654,12 +604,12 @@ class Client(object):
 
     ### LIST COMMANDS
     def blpop(self, keys, timeout=0, callback=None):
-        tokens = list(keys)
+        tokens = to_list(keys)
         tokens.append(timeout)
         self.execute_command('BLPOP', *tokens, callback=callback)
 
     def brpop(self, keys, timeout=0, callback=None):
-        tokens = list(keys)
+        tokens = to_list(keys)
         tokens.append(timeout)
         self.execute_command('BRPOP', *tokens, callback=callback)
 
@@ -990,6 +940,7 @@ class Client(object):
 
 
 class Pipeline(Client):
+
     def __init__(self, transactional, *args, **kwargs):
         super(Pipeline, self).__init__(*args, **kwargs)
         self.transactional = transactional
@@ -1016,6 +967,10 @@ class Pipeline(Client):
                 results.append(e)
         return results
 
+    def format_pipeline_request(self, command_stack):
+        return ''.join(self.format_command(c.cmd, *c.args, **c.kwargs)
+                       for c in command_stack)
+
     @gen.engine
     def execute(self, callback=None):
         command_stack = self.command_stack
@@ -1026,7 +981,7 @@ class Pipeline(Client):
                           + command_stack \
                           + [CmdLine('EXEC')]
 
-        request = format_pipeline_request(command_stack)
+        request = self.format_pipeline_request(command_stack)
 
         if not self.connection.ready():
             yield gen.Task(self.connection.wait_until_ready)
@@ -1054,7 +1009,8 @@ class Pipeline(Client):
                 try:
                     cmd_line = cmds.next()
                     if self.transactional and cmd_line.cmd != 'EXEC':
-                        response = yield gen.Task(self.process_data, data,
+                        response = yield gen.Task(self.process_data,
+                                                  data,
                                                   CmdLine('MULTI_PART'))
                     else:
                         response = yield gen.Task(self.process_data,
